package io.tswf.selenium.undetected.chrome.driver.factory

import groovy.transform.CompileStatic
import io.tswf.selenium.undetected.chrome.driver.UndetectedChromeDriver
import io.tswf.selenium.undetected.chrome.driver.util.SocketUtil
import io.tswf.selenium.undetected.chrome.patcher.ChromeDriverUndetectedPatcher
import io.tswf.selenium.undetected.chrome.patcher.ChromeDriverUndetectedPatcherImpl
import io.tswf.selenium.undetected.chrome.process.ChromeBrowserProcessStarter
import io.tswf.selenium.undetected.chrome.settings.BufferedChromeOptions
import org.openqa.selenium.chrome.ChromeDriver
import org.openqa.selenium.chrome.ChromeOptions

import java.nio.file.Files
import java.util.function.Consumer

@CompileStatic
interface UndetectedChromeDriverBuilder {
    UndetectedChromeDriverBuilder withConfiguration(
            @DelegatesTo(ChromeDriver)
            Consumer<ChromeDriver> configurer
    )

    UndetectedChromeDriverBuilder withOptions(
            @DelegatesTo(ChromeOptions)
            Consumer<ChromeOptions> configurer
    )

    UndetectedChromeDriverBuilder withUserDir(File userDir)

    UndetectedChromeDriverBuilder withHeadlessUi()

    UndetectedChromeDriverBuilder withMutedAudio()

    UndetectedChromeDriverBuilder withUndetectedPatcher()

    UndetectedChromeDriverBuilder withBrowserResolution(int width, int height)

    UndetectedChromeDriverBuilder withUndetectedPatcher(ChromeDriverUndetectedPatcher patcher)

    ChromeDriver build()
}

@CompileStatic
class UndetectedChromeDriverBuilderImpl implements UndetectedChromeDriverBuilder {

    BufferedChromeOptions chromeOptions = new BufferedChromeOptions()

    File userDataDir
    ChromeDriverUndetectedPatcher driverUndetectedPatcher
    Tuple2<Integer, Integer> browserResolution
    boolean headless

    Consumer<ChromeDriver> postCreationHandler

    File driverExecutable
    File chromeExecutable

    ChromeBrowserProcessStarter browserProcessStarter

    UndetectedChromeDriverBuilderImpl(File driverExecutable, File chromeExecutable, ChromeBrowserProcessStarter browserProcessStarter) {
        this.driverExecutable = driverExecutable
        this.chromeExecutable = chromeExecutable
        this.browserProcessStarter = browserProcessStarter

        chromeOptions.addArguments("--no-default-browser-check", "--no-first-run")
    }

    @Override
    UndetectedChromeDriverBuilder withConfiguration(@DelegatesTo(ChromeDriver) Consumer<ChromeDriver> configurer) {
        postCreationHandler = configurer
        return this
    }

    @Override
    UndetectedChromeDriverBuilder withOptions(@DelegatesTo(ChromeOptions) Consumer<ChromeOptions> configurer) {
        configurer.accept(chromeOptions)
        return this
    }

    @Override
    UndetectedChromeDriverBuilder withUserDir(File userDir) {
        this.userDataDir = userDir
        return this
    }

    @Override
    UndetectedChromeDriverBuilder withHeadlessUi() {
        headless = true
        chromeOptions.addArguments("--headless=new")
        return this
    }

    @Override
    UndetectedChromeDriverBuilder withMutedAudio() {
        chromeOptions.addArguments("--mute-audio")
        return this
    }

    @Override
    UndetectedChromeDriverBuilder withUndetectedPatcher(ChromeDriverUndetectedPatcher patcher = new ChromeDriverUndetectedPatcherImpl()) {
        this.driverUndetectedPatcher = patcher
        return this
    }

    @Override
    ChromeDriver build() {
        chromeOptions.binary = chromeExecutable

        if (userDataDir == null) {
            userDataDir = Files.createTempDirectory("selenium-chrome-session").toFile()

            addShutdownHook {
                retrying(5, 500, { it.first() }) {
                    userDataDir.deleteIfExists()
                }
            }
        }

        if (userDataDir != null) {
            userDataDir.mkdirsUnsafe()
            chromeOptions.addArguments("--user-data-dir="+userDataDir.absoluteCanonicalFile.path)
        }

        if (driverUndetectedPatcher != null) {
            if (!driverUndetectedPatcher.isDriverExecutablePatched(driverExecutable)) {
                driverUndetectedPatcher.patchDriverExecutable(driverExecutable)
            }
        }

        if (browserResolution != null) {
            chromeOptions.addArguments("--window-size=${browserResolution.v1},${browserResolution.v2}")
        }

        // Подключение к нашему процессу chrome драйвер будет выполнять по http. Для этого ему нужно указать хост и порт, на которых он будет работать. Эти же 2 флага будет использовать chrome.
        chromeOptions.addArguments("--remote-debugging-host=127.0.0.1", "--remote-debugging-port=${SocketUtil.findFreeTcpPort()}")

        def remoteChromeBrowserProcess = startRemoteChromeBrowserProcess()

        chromeOptions.flushArguments()

        def newChromeDriver = new UndetectedChromeDriver(chromeOptions, headless, userDataDir, remoteChromeBrowserProcess)
        postCreationHandler?.accept(newChromeDriver)

        return newChromeDriver
    }

    @Override
    UndetectedChromeDriverBuilder withBrowserResolution(int width, int height) {
        browserResolution = new Tuple2<>(width, height)
        return this
    }

    /**
     * Запускаем демона Chrome сами, потому как если это сделает браузер - он будет детектиться как бот.
     * Для подключения к браузеру будет использоваться удаленный дебаг.
     * @return Запущенный процесс chrome
     */
    Process startRemoteChromeBrowserProcess() {
        def chromeArgs = chromeOptions.argumentsBuffer.toList()

        def remoteChromeBrowserProcess = browserProcessStarter.startChromeBrowserProcess(chromeExecutable, chromeArgs)
        return remoteChromeBrowserProcess
    }
}